"""OpenClaw-style browser tool for MindRoom.

This exposes a single ``browser`` function with an ``action`` parameter,
matching OpenClaw's tool surface at a high level.
"""
# ruff: noqa: N803, A002

from __future__ import annotations

import asyncio
import json
import shutil
from dataclasses import dataclass, field
from pathlib import Path
from typing import TYPE_CHECKING, Any
from uuid import uuid4

from agno.tools import Toolkit
from playwright.async_api import Browser, BrowserContext, ConsoleMessage, Dialog, Page, Playwright, async_playwright

from mindroom.tool_system.runtime_context import get_tool_runtime_context

if TYPE_CHECKING:
    from mindroom.constants import RuntimePaths

_DEFAULT_PROFILE = "openclaw"
_DEFAULT_SNAPSHOT_LIMIT = 200
_DEFAULT_AI_SNAPSHOT_MAX_CHARS = 12_000
_DEFAULT_TIMEOUT_MS = 30_000
_MAX_CONSOLE_ENTRIES = 200
_VIEWPORT_WIDTH = 1280
_VIEWPORT_HEIGHT = 720

_BROWSER_ACTIONS = {
    "status",
    "start",
    "stop",
    "profiles",
    "tabs",
    "open",
    "focus",
    "close",
    "snapshot",
    "screenshot",
    "navigate",
    "console",
    "pdf",
    "upload",
    "dialog",
    "act",
}

_SNAPSHOT_JS = """
({ selector, limit, depth, interactiveOnly }) => {
  const root = selector ? document.querySelector(selector) : (document.body || document.documentElement);
  if (!root) return [];

  const maxItems = Math.max(1, Number.isFinite(limit) ? Number(limit) : 200);
  const maxDepth = Math.max(1, Number.isFinite(depth) ? Number(depth) : 12);

  const isVisible = (el) => {
    if (!(el instanceof HTMLElement)) return false;
    const style = window.getComputedStyle(el);
    if (style.display === "none" || style.visibility === "hidden" || style.opacity === "0") return false;
    const rect = el.getBoundingClientRect();
    return rect.width > 0 && rect.height > 0;
  };

  const isInteractive = (el) => {
    const tag = el.tagName.toLowerCase();
    if (["a", "button", "input", "select", "textarea", "summary", "option", "label"].includes(tag)) return true;
    if (el.hasAttribute("contenteditable")) return true;
    if (el.hasAttribute("onclick")) return true;
    const role = (el.getAttribute("role") || "").trim();
    return role.length > 0;
  };

  const inferRole = (el) => {
    const explicit = (el.getAttribute("role") || "").trim();
    if (explicit) return explicit;
    const tag = el.tagName.toLowerCase();
    if (tag === "a") return "link";
    if (tag === "button") return "button";
    if (tag === "input") return "input";
    if (tag === "select") return "select";
    if (tag === "textarea") return "textbox";
    return "";
  };

  const selectorFor = (el) => {
    if (el.id) return `#${CSS.escape(el.id)}`;
    const parts = [];
    let node = el;
    let safety = 0;
    while (node && node.nodeType === Node.ELEMENT_NODE && safety < 12) {
      let part = node.tagName.toLowerCase();
      const parent = node.parentElement;
      if (parent) {
        const siblings = Array.from(parent.children).filter((sib) => sib.tagName === node.tagName);
        if (siblings.length > 1) {
          const idx = siblings.indexOf(node) + 1;
          part += `:nth-of-type(${idx})`;
        }
      }
      parts.unshift(part);
      if (node.id) {
        parts[0] = `#${CSS.escape(node.id)}`;
        break;
      }
      node = parent;
      safety += 1;
    }
    return parts.join(" > ");
  };

  const depthFromRoot = (el) => {
    let d = 0;
    let node = el;
    while (node && node !== root) {
      d += 1;
      node = node.parentElement;
      if (d > 200) break;
    }
    return d;
  };

  const candidates = [root, ...root.querySelectorAll("*")];
  const rows = [];
  for (const el of candidates) {
    if (!(el instanceof HTMLElement)) continue;
    if (!isVisible(el)) continue;
    if (depthFromRoot(el) > maxDepth) continue;
    if (interactiveOnly && !isInteractive(el)) continue;
    const text = (
      el.getAttribute("aria-label") ||
      el.getAttribute("alt") ||
      el.getAttribute("title") ||
      el.innerText ||
      el.textContent ||
      el.getAttribute("value") ||
      ""
    ).replace(/\\s+/g, " ").trim();
    rows.push({
      selector: selectorFor(el),
      role: inferRole(el),
      name: text.slice(0, 140),
      tag: el.tagName.toLowerCase(),
      type: (el.getAttribute("type") || "").trim(),
    });
    if (rows.length >= maxItems) break;
  }
  return rows;
}
"""


@dataclass
class _BrowserTabState:
    """State for one browser tab."""

    target_id: str
    page: Page
    refs: dict[str, str] = field(default_factory=dict)
    pending_dialog: dict[str, Any] | None = None
    console: list[dict[str, Any]] = field(default_factory=list)


@dataclass
class _BrowserProfileState:
    """State for one browser profile."""

    playwright: Playwright
    browser: Browser
    context: BrowserContext
    tabs: dict[str, _BrowserTabState] = field(default_factory=dict)
    active_target_id: str | None = None


def _clean_str(value: object) -> str | None:
    """Normalize a value to a stripped string."""
    if not isinstance(value, str):
        return None
    cleaned = value.strip()
    return cleaned if cleaned else None


class BrowserTools(Toolkit):
    """OpenClaw-style browser control for MindRoom agents."""

    def __init__(self, runtime_paths: RuntimePaths, *, output_dir: Path | str | None = None) -> None:
        super().__init__(name="browser", tools=[self.browser])
        self._runtime_paths = runtime_paths
        self._profiles: dict[str, _BrowserProfileState] = {}
        self._lock = asyncio.Lock()
        self._output_dir = Path(output_dir).expanduser().resolve() if output_dir is not None else None
        if self._output_dir is not None:
            self._output_dir.mkdir(parents=True, exist_ok=True)
        self._close_task: asyncio.Task[None] | None = None

    async def _close_profiles(self) -> None:
        """Close all active browser profiles."""
        for profile_name in list(self._profiles.keys()):
            await self._stop_profile(profile_name)

    def close(self) -> None:
        """Close toolkit resources."""
        try:
            loop = asyncio.get_running_loop()
        except RuntimeError:
            asyncio.run(self._close_profiles())
            return
        self._close_task = loop.create_task(self._close_profiles())

    async def browser(  # noqa: C901, PLR0911, PLR0912
        self,
        action: str,
        target: str | None = None,
        node: str | None = None,
        profile: str | None = None,
        targetUrl: str | None = None,
        targetId: str | None = None,
        limit: int | None = None,
        maxChars: int | None = None,
        mode: str | None = None,
        snapshotFormat: str | None = None,
        refs: str | None = None,
        interactive: bool | None = None,
        compact: bool | None = None,
        depth: int | None = None,
        selector: str | None = None,
        frame: str | None = None,
        labels: bool | None = None,
        fullPage: bool | None = None,
        ref: str | None = None,
        element: str | None = None,
        type: str | None = None,
        level: str | None = None,
        paths: list[str] | None = None,
        inputRef: str | None = None,
        timeoutMs: int | None = None,
        accept: bool | None = None,
        promptText: str | None = None,
        request: dict[str, Any] | None = None,
    ) -> str:
        """Control browser state and actions.

        Args:
            action: Browser action (status/start/stop/profiles/tabs/open/focus/close/snapshot/screenshot/navigate/console/pdf/upload/dialog/act)
            target: Browser target location. Only ``host`` is currently supported.
            node: Node id (OpenClaw compatibility field; unsupported in MindRoom runtime).
            profile: Browser profile name (defaults to ``openclaw``).
            targetUrl: URL for ``open`` and ``navigate`` actions.
            targetId: Tab target id for actions that address a specific tab.
            limit: Snapshot item limit.
            maxChars: Snapshot text limit.
            mode: Snapshot mode (supports ``efficient``).
            snapshotFormat: ``ai`` or ``aria``.
            refs: Snapshot refs mode (accepted for compatibility).
            interactive: Snapshot interactive filtering.
            compact: Snapshot compact mode hint.
            depth: Snapshot traversal depth.
            selector: Root selector for snapshot.
            frame: Frame selector (accepted for compatibility; ignored for now).
            labels: Snapshot label hint (accepted for compatibility; ignored for now).
            fullPage: Full-page capture for screenshots.
            ref: Snapshot ref id or CSS selector.
            element: CSS selector for element-specific actions.
            type: Screenshot type (``png`` or ``jpeg``).
            level: Console log level filter.
            paths: Upload file paths.
            inputRef: Upload input selector or ref.
            timeoutMs: Timeout for wait/upload/dialog actions.
            accept: Whether to accept dialog.
            promptText: Prompt text for dialog accept.
            request: Act request object.

        Returns:
            JSON-encoded result payload.

        """
        normalized_action = action.strip().lower()
        if normalized_action not in _BROWSER_ACTIONS:
            msg = f"Unknown action: {action}"
            raise ValueError(msg)

        self._validate_target(target=target, node=node)
        profile_name = _clean_str(profile) or _DEFAULT_PROFILE

        if normalized_action == "status":
            return json.dumps(await self._status_payload(profile_name), sort_keys=True)
        if normalized_action == "start":
            state = await self._ensure_profile(profile_name)
            payload = await self._profile_status(profile_name, state)
            payload["action"] = "start"
            return json.dumps(payload, sort_keys=True)
        if normalized_action == "stop":
            await self._stop_profile(profile_name)
            return json.dumps({"action": "stop", "profile": profile_name, "status": "ok"}, sort_keys=True)
        if normalized_action == "profiles":
            return json.dumps(await self._profiles_payload(profile_name), sort_keys=True)
        if normalized_action == "tabs":
            state = await self._ensure_profile(profile_name)
            return json.dumps(await self._tabs_payload(profile_name, state), sort_keys=True)
        if normalized_action == "open":
            target_url = _clean_str(targetUrl)
            if target_url is None:
                msg = "targetUrl required for action=open"
                raise ValueError(msg)
            return json.dumps(await self._open_tab(profile_name, target_url), sort_keys=True)
        if normalized_action == "focus":
            target_id = _clean_str(targetId)
            if target_id is None:
                msg = "targetId required for action=focus"
                raise ValueError(msg)
            return json.dumps(await self._focus_tab(profile_name, target_id), sort_keys=True)
        if normalized_action == "close":
            return json.dumps(await self._close_tab(profile_name, _clean_str(targetId)), sort_keys=True)
        if normalized_action == "snapshot":
            return json.dumps(
                await self._snapshot(
                    profile_name=profile_name,
                    target_id=_clean_str(targetId),
                    snapshot_format=_clean_str(snapshotFormat),
                    limit=limit,
                    max_chars=maxChars,
                    mode=_clean_str(mode),
                    refs_mode=_clean_str(refs),
                    interactive=interactive,
                    compact=compact,
                    depth=depth,
                    selector=_clean_str(selector),
                    frame=_clean_str(frame),
                    labels=labels,
                ),
                sort_keys=True,
            )
        if normalized_action == "screenshot":
            return json.dumps(
                await self._screenshot(
                    profile_name=profile_name,
                    target_id=_clean_str(targetId),
                    full_page=bool(fullPage),
                    ref=_clean_str(ref),
                    element=_clean_str(element),
                    image_type=_clean_str(type),
                ),
                sort_keys=True,
            )
        if normalized_action == "navigate":
            target_url = _clean_str(targetUrl)
            if target_url is None:
                msg = "targetUrl required for action=navigate"
                raise ValueError(msg)
            return json.dumps(
                await self._navigate(profile_name, target_url, _clean_str(targetId)),
                sort_keys=True,
            )
        if normalized_action == "console":
            return json.dumps(
                await self._console(
                    profile_name=profile_name,
                    target_id=_clean_str(targetId),
                    level=_clean_str(level),
                ),
                sort_keys=True,
            )
        if normalized_action == "pdf":
            return json.dumps(await self._pdf(profile_name, _clean_str(targetId)), sort_keys=True)
        if normalized_action == "upload":
            if not paths:
                msg = "paths required for action=upload"
                raise ValueError(msg)
            return json.dumps(
                await self._upload(
                    profile_name=profile_name,
                    target_id=_clean_str(targetId),
                    paths=paths,
                    ref=_clean_str(ref),
                    input_ref=_clean_str(inputRef),
                    element=_clean_str(element),
                    timeout_ms=timeoutMs,
                ),
                sort_keys=True,
            )
        if normalized_action == "dialog":
            return json.dumps(
                await self._dialog(
                    profile_name=profile_name,
                    target_id=_clean_str(targetId),
                    accept=bool(accept),
                    prompt_text=_clean_str(promptText),
                    timeout_ms=timeoutMs,
                ),
                sort_keys=True,
            )
        if normalized_action == "act":
            if not isinstance(request, dict):
                msg = "request required for action=act"
                raise ValueError(msg)
            return json.dumps(
                await self._act(
                    profile_name=profile_name,
                    request=request,
                    fallback_target_id=_clean_str(targetId),
                ),
                sort_keys=True,
            )

        msg = f"Unhandled action: {action}"
        raise ValueError(msg)

    @staticmethod
    def _validate_target(*, target: str | None, node: str | None) -> None:
        normalized_target = _clean_str(target)
        if node is not None and normalized_target not in {None, "node"}:
            msg = "node parameter is not supported in MindRoom."
            raise ValueError(msg)
        if normalized_target in {"sandbox", "node"} or node is not None:
            msg = "MindRoom browser tool currently supports host target only."
            raise ValueError(msg)
        if normalized_target not in {None, "host"}:
            msg = f"Unsupported target: {target}"
            raise ValueError(msg)

    async def _status_payload(self, profile_name: str) -> dict[str, Any]:
        async with self._lock:
            state = self._profiles.get(profile_name)
            if state is None:
                return {"action": "status", "profile": profile_name, "running": False, "status": "ok", "tabs": []}
            return await self._profile_status(profile_name, state)

    async def _profiles_payload(self, selected_profile: str) -> dict[str, Any]:
        async with self._lock:
            running = sorted(self._profiles.keys())
        advertised = sorted({_DEFAULT_PROFILE, "chrome", *running})
        return {
            "action": "profiles",
            "profiles": advertised,
            "running_profiles": running,
            "selected_profile": selected_profile,
            "status": "ok",
        }

    async def _profile_status(self, profile_name: str, state: _BrowserProfileState) -> dict[str, Any]:
        tabs = await self._tab_list(state)
        return {
            "action": "status",
            "activeTargetId": state.active_target_id,
            "profile": profile_name,
            "running": True,
            "status": "ok",
            "tabCount": len(tabs),
            "tabs": tabs,
        }

    async def _tab_list(self, state: _BrowserProfileState) -> list[dict[str, Any]]:
        payload_tabs: list[dict[str, Any]] = []
        stale: list[str] = []
        for target_id, tab in state.tabs.items():
            if tab.page.is_closed():
                stale.append(target_id)
                continue
            title = await tab.page.title()
            payload_tabs.append(
                {
                    "active": target_id == state.active_target_id,
                    "targetId": target_id,
                    "title": title,
                    "url": tab.page.url,
                },
            )
        for target_id in stale:
            self._remove_tab(state, target_id)
        return payload_tabs

    async def _tabs_payload(self, profile_name: str, state: _BrowserProfileState) -> dict[str, Any]:
        tabs = await self._tab_list(state)
        return {
            "action": "tabs",
            "activeTargetId": state.active_target_id,
            "profile": profile_name,
            "status": "ok",
            "tabs": tabs,
        }

    async def _open_tab(self, profile_name: str, target_url: str) -> dict[str, Any]:
        state = await self._ensure_profile(profile_name)
        page = await state.context.new_page()
        target_id = self._register_tab(state, page)
        await page.goto(target_url, wait_until="domcontentloaded", timeout=_DEFAULT_TIMEOUT_MS)
        state.active_target_id = target_id
        return {
            "action": "open",
            "profile": profile_name,
            "status": "ok",
            "targetId": target_id,
            "title": await page.title(),
            "url": page.url,
        }

    async def _focus_tab(self, profile_name: str, target_id: str) -> dict[str, Any]:
        state = await self._ensure_profile(profile_name)
        if target_id not in state.tabs or state.tabs[target_id].page.is_closed():
            msg = f"tab not found: {target_id}"
            raise ValueError(msg)
        state.active_target_id = target_id
        page = state.tabs[target_id].page
        return {
            "action": "focus",
            "profile": profile_name,
            "status": "ok",
            "targetId": target_id,
            "title": await page.title(),
            "url": page.url,
        }

    async def _close_tab(self, profile_name: str, target_id: str | None) -> dict[str, Any]:
        state = await self._ensure_profile(profile_name)
        resolved_target_id, tab = await self._resolve_tab(state, target_id)
        await tab.page.close()
        self._remove_tab(state, resolved_target_id)
        return {
            "action": "close",
            "profile": profile_name,
            "status": "ok",
            "targetId": resolved_target_id,
        }

    async def _navigate(self, profile_name: str, target_url: str, target_id: str | None) -> dict[str, Any]:
        state = await self._ensure_profile(profile_name)
        resolved_target_id, tab = await self._resolve_tab(state, target_id)
        await tab.page.goto(target_url, wait_until="domcontentloaded", timeout=_DEFAULT_TIMEOUT_MS)
        state.active_target_id = resolved_target_id
        return {
            "action": "navigate",
            "profile": profile_name,
            "status": "ok",
            "targetId": resolved_target_id,
            "title": await tab.page.title(),
            "url": tab.page.url,
        }

    async def _console(self, profile_name: str, target_id: str | None, level: str | None) -> dict[str, Any]:
        state = await self._ensure_profile(profile_name)
        if target_id is not None:
            _, tab = await self._resolve_tab(state, target_id)
            entries = tab.console
        else:
            entries = [entry for tab in state.tabs.values() for entry in tab.console]
        if level is not None:
            entries = [entry for entry in entries if entry.get("level") == level]
        return {
            "action": "console",
            "entries": entries[-_MAX_CONSOLE_ENTRIES:],
            "level": level,
            "profile": profile_name,
            "status": "ok",
            "targetId": target_id,
        }

    async def _pdf(self, profile_name: str, target_id: str | None) -> dict[str, Any]:
        state = await self._ensure_profile(profile_name)
        resolved_target_id, tab = await self._resolve_tab(state, target_id)
        output_path = self._next_output_path("pdf")
        await tab.page.pdf(path=str(output_path))
        return {
            "action": "pdf",
            "path": str(output_path),
            "profile": profile_name,
            "status": "ok",
            "targetId": resolved_target_id,
        }

    async def _upload(
        self,
        *,
        profile_name: str,
        target_id: str | None,
        paths: list[str],
        ref: str | None,
        input_ref: str | None,
        element: str | None,
        timeout_ms: int | None,
    ) -> dict[str, Any]:
        state = await self._ensure_profile(profile_name)
        resolved_target_id, tab = await self._resolve_tab(state, target_id)
        selector = self._resolve_selector(tab, input_ref or ref or element)
        if selector is None:
            msg = "upload requires inputRef, ref, or element"
            raise ValueError(msg)
        normalized_paths = [str(Path(path).expanduser()) for path in paths]
        locator = tab.page.locator(selector).first
        await locator.set_input_files(normalized_paths, timeout=timeout_ms or _DEFAULT_TIMEOUT_MS)
        return {
            "action": "upload",
            "paths": normalized_paths,
            "profile": profile_name,
            "selector": selector,
            "status": "ok",
            "targetId": resolved_target_id,
        }

    async def _dialog(
        self,
        *,
        profile_name: str,
        target_id: str | None,
        accept: bool,
        prompt_text: str | None,
        timeout_ms: int | None,
    ) -> dict[str, Any]:
        state = await self._ensure_profile(profile_name)
        resolved_target_id, tab = await self._resolve_tab(state, target_id)
        tab.pending_dialog = {
            "accept": accept,
            "promptText": prompt_text,
            "timeoutMs": timeout_ms,
        }
        return {
            "accept": accept,
            "action": "dialog",
            "armed": True,
            "profile": profile_name,
            "promptText": prompt_text,
            "status": "ok",
            "targetId": resolved_target_id,
            "timeoutMs": timeout_ms,
        }

    async def _screenshot(
        self,
        *,
        profile_name: str,
        target_id: str | None,
        full_page: bool,
        ref: str | None,
        element: str | None,
        image_type: str | None,
    ) -> dict[str, Any]:
        state = await self._ensure_profile(profile_name)
        resolved_target_id, tab = await self._resolve_tab(state, target_id)
        resolved_type = "jpeg" if image_type == "jpeg" else "png"
        output_path = self._next_output_path("jpg" if resolved_type == "jpeg" else "png")
        selector = self._resolve_selector(tab, element or ref)
        if selector is None:
            await tab.page.screenshot(path=str(output_path), type=resolved_type, full_page=full_page)
        else:
            await tab.page.locator(selector).first.screenshot(path=str(output_path), type=resolved_type)
        return {
            "action": "screenshot",
            "fullPage": full_page,
            "path": str(output_path),
            "profile": profile_name,
            "selector": selector,
            "status": "ok",
            "targetId": resolved_target_id,
            "type": resolved_type,
        }

    async def _snapshot(
        self,
        *,
        profile_name: str,
        target_id: str | None,
        snapshot_format: str | None,
        limit: int | None,
        max_chars: int | None,
        mode: str | None,
        refs_mode: str | None,
        interactive: bool | None,
        compact: bool | None,
        depth: int | None,
        selector: str | None,
        frame: str | None,
        labels: bool | None,
    ) -> dict[str, Any]:
        state = await self._ensure_profile(profile_name)
        resolved_target_id, tab = await self._resolve_tab(state, target_id)
        if frame is not None:
            del frame  # Not implemented in local Playwright path yet.
        if labels is not None:
            del labels  # Label overlays are currently not implemented.

        fmt = "aria" if snapshot_format == "aria" else "ai"
        resolved_limit = max(1, limit) if isinstance(limit, int) and limit > 0 else _DEFAULT_SNAPSHOT_LIMIT
        resolved_depth = max(1, depth) if isinstance(depth, int) and depth > 0 else 12
        interactive_only = interactive if isinstance(interactive, bool) else True
        rows = await tab.page.evaluate(
            _SNAPSHOT_JS,
            {
                "depth": resolved_depth,
                "interactiveOnly": interactive_only,
                "limit": resolved_limit,
                "selector": selector,
            },
        )
        if not isinstance(rows, list):
            rows = []

        ref_entries: list[dict[str, Any]] = []
        tab.refs = {}
        for index, row in enumerate(rows, start=1):
            if not isinstance(row, dict):
                continue
            selector_value = _clean_str(row.get("selector"))
            if selector_value is None:
                continue
            ref_id = f"e{index}"
            tab.refs[ref_id] = selector_value
            ref_entries.append(
                {
                    "name": _clean_str(row.get("name")) or "",
                    "ref": ref_id,
                    "role": _clean_str(row.get("role")) or "",
                    "selector": selector_value,
                    "tag": _clean_str(row.get("tag")) or "",
                    "type": _clean_str(row.get("type")) or "",
                },
            )

        state.active_target_id = resolved_target_id
        title = await tab.page.title()
        if fmt == "aria":
            return {
                "action": "snapshot",
                "count": len(ref_entries),
                "format": "aria",
                "interactive": interactive_only,
                "mode": mode,
                "profile": profile_name,
                "refs": refs_mode or "role",
                "status": "ok",
                "targetId": resolved_target_id,
                "title": title,
                "url": tab.page.url,
                "elements": ref_entries,
            }

        lines: list[str] = [f"URL: {tab.page.url}", f"Title: {title}", "Elements:"]
        for entry in ref_entries:
            role_or_tag = entry["role"] or entry["tag"] or "element"
            name = entry["name"]
            base = f"[{entry['ref']}] {role_or_tag}"
            lines.append(f"{base}: {name}" if name else base)
        snapshot_text = "\n".join(lines if not compact else lines[:2] + lines[3:])
        resolved_max_chars = self._resolve_max_chars(max_chars=max_chars, mode=mode)
        if resolved_max_chars is not None and len(snapshot_text) > resolved_max_chars:
            snapshot_text = snapshot_text[:resolved_max_chars].rstrip() + "\n…"

        return {
            "action": "snapshot",
            "count": len(ref_entries),
            "format": "ai",
            "interactive": interactive_only,
            "mode": mode,
            "profile": profile_name,
            "refs": refs_mode or "role",
            "snapshot": snapshot_text,
            "status": "ok",
            "targetId": resolved_target_id,
            "title": title,
            "url": tab.page.url,
        }

    @staticmethod
    def _resolve_max_chars(*, max_chars: int | None, mode: str | None) -> int | None:
        if isinstance(max_chars, int):
            return max_chars if max_chars > 0 else None
        if mode == "efficient":
            return None
        return _DEFAULT_AI_SNAPSHOT_MAX_CHARS

    async def _act(  # noqa: C901, PLR0911, PLR0912, PLR0915
        self,
        *,
        profile_name: str,
        request: dict[str, Any],
        fallback_target_id: str | None,
    ) -> dict[str, Any]:
        state = await self._ensure_profile(profile_name)
        requested_target_id = _clean_str(request.get("targetId")) or fallback_target_id
        resolved_target_id, tab = await self._resolve_tab(state, requested_target_id)
        kind = _clean_str(request.get("kind"))
        if kind is None:
            msg = "request.kind required"
            raise ValueError(msg)

        if kind == "click":
            selector = self._resolve_selector(tab, _clean_str(request.get("ref")))
            if selector is None:
                msg = "click requires request.ref"
                raise ValueError(msg)
            click_kwargs: dict[str, Any] = {}
            if request.get("doubleClick") is True:
                click_kwargs["click_count"] = 2
            button = _clean_str(request.get("button"))
            if button is not None:
                click_kwargs["button"] = button
            modifiers = request.get("modifiers")
            if isinstance(modifiers, list):
                click_kwargs["modifiers"] = [str(modifier) for modifier in modifiers]
            await tab.page.locator(selector).first.click(**click_kwargs)
            return self._act_result(profile_name, resolved_target_id, kind)

        if kind == "type":
            selector = self._resolve_selector(tab, _clean_str(request.get("ref")))
            if selector is None:
                msg = "type requires request.ref"
                raise ValueError(msg)
            text = str(request.get("text", ""))
            locator = tab.page.locator(selector).first
            if request.get("slowly") is True:
                await locator.type(text, delay=75)
            else:
                await locator.fill(text)
            if request.get("submit") is True:
                await locator.press("Enter")
            return self._act_result(profile_name, resolved_target_id, kind, text=text)

        if kind == "press":
            key = _clean_str(request.get("key"))
            if key is None:
                msg = "press requires request.key"
                raise ValueError(msg)
            await tab.page.keyboard.press(key)
            return self._act_result(profile_name, resolved_target_id, kind, key=key)

        if kind == "hover":
            selector = self._resolve_selector(tab, _clean_str(request.get("ref")))
            if selector is None:
                msg = "hover requires request.ref"
                raise ValueError(msg)
            await tab.page.locator(selector).first.hover()
            return self._act_result(profile_name, resolved_target_id, kind)

        if kind == "drag":
            start_selector = self._resolve_selector(tab, _clean_str(request.get("startRef")))
            end_selector = self._resolve_selector(tab, _clean_str(request.get("endRef")))
            if start_selector is None or end_selector is None:
                msg = "drag requires request.startRef and request.endRef"
                raise ValueError(msg)
            await tab.page.drag_and_drop(start_selector, end_selector)
            return self._act_result(profile_name, resolved_target_id, kind)

        if kind == "select":
            selector = self._resolve_selector(tab, _clean_str(request.get("ref")))
            values = request.get("values")
            if selector is None or not isinstance(values, list) or not values:
                msg = "select requires request.ref and request.values"
                raise ValueError(msg)
            await tab.page.select_option(selector, [str(value) for value in values])
            return self._act_result(profile_name, resolved_target_id, kind, values=values)

        if kind == "fill":
            fields = request.get("fields")
            if not isinstance(fields, list) or not fields:
                msg = "fill requires request.fields"
                raise ValueError(msg)
            updated: list[dict[str, str]] = []
            for field in fields:
                if not isinstance(field, dict):
                    continue
                selector = self._resolve_selector(
                    tab,
                    _clean_str(field.get("ref")) or _clean_str(field.get("selector")),
                )
                if selector is None:
                    continue
                value = str(field.get("value", ""))
                await tab.page.locator(selector).first.fill(value)
                updated.append({"selector": selector, "value": value})
            if not updated:
                msg = "fill requires at least one field with a valid ref or selector"
                raise ValueError(msg)
            return self._act_result(profile_name, resolved_target_id, kind, fields=updated)

        if kind == "resize":
            width = request.get("width")
            height = request.get("height")
            if not isinstance(width, int) or not isinstance(height, int) or width <= 0 or height <= 0:
                msg = "resize requires positive integer request.width and request.height"
                raise ValueError(msg)
            await tab.page.set_viewport_size({"height": height, "width": width})
            return self._act_result(profile_name, resolved_target_id, kind, width=width, height=height)

        if kind == "wait":
            time_ms = request.get("timeMs")
            text = _clean_str(request.get("text"))
            text_gone = _clean_str(request.get("textGone"))
            if isinstance(time_ms, int) and time_ms >= 0:
                await tab.page.wait_for_timeout(time_ms)
            elif text is not None:
                await tab.page.wait_for_selector(f"text={text}", state="visible", timeout=_DEFAULT_TIMEOUT_MS)
            elif text_gone is not None:
                await tab.page.wait_for_selector(f"text={text_gone}", state="detached", timeout=_DEFAULT_TIMEOUT_MS)
            else:
                await tab.page.wait_for_timeout(500)
            return self._act_result(profile_name, resolved_target_id, kind)

        if kind == "evaluate":
            fn = _clean_str(request.get("fn"))
            if fn is None:
                msg = "evaluate requires request.fn"
                raise ValueError(msg)
            selector = self._resolve_selector(tab, _clean_str(request.get("ref")))
            result = await tab.page.evaluate(fn) if selector is None else await tab.page.eval_on_selector(selector, fn)
            return self._act_result(profile_name, resolved_target_id, kind, result=result)

        if kind == "close":
            await tab.page.close()
            self._remove_tab(state, resolved_target_id)
            return self._act_result(profile_name, resolved_target_id, kind)

        msg = f"Unsupported act kind: {kind}"
        raise ValueError(msg)

    @staticmethod
    def _act_result(profile_name: str, target_id: str, kind: str, **extra: object) -> dict[str, Any]:
        payload: dict[str, Any] = {
            "action": "act",
            "kind": kind,
            "profile": profile_name,
            "status": "ok",
            "targetId": target_id,
        }
        payload.update(extra)
        return payload

    async def _ensure_profile(self, profile_name: str) -> _BrowserProfileState:
        async with self._lock:
            state = self._profiles.get(profile_name)
            if state is not None:
                return state

            playwright = await async_playwright().start()
            executable = (
                self._runtime_paths.env_value("BROWSER_EXECUTABLE_PATH")
                or shutil.which("chromium")
                or shutil.which("google-chrome-stable")
            )
            launch_kwargs: dict[str, Any] = {"headless": True}
            if executable:
                launch_kwargs["executable_path"] = executable
            browser = await playwright.chromium.launch(**launch_kwargs)
            context = await browser.new_context(viewport={"height": _VIEWPORT_HEIGHT, "width": _VIEWPORT_WIDTH})
            state = _BrowserProfileState(playwright=playwright, browser=browser, context=context)
            self._profiles[profile_name] = state

            page = await context.new_page()
            target_id = self._register_tab(state, page)
            state.active_target_id = target_id
            return state

    async def _stop_profile(self, profile_name: str) -> None:
        async with self._lock:
            state = self._profiles.pop(profile_name, None)
        if state is None:
            return
        await state.context.close()
        await state.browser.close()
        await state.playwright.stop()

    async def _resolve_tab(
        self,
        state: _BrowserProfileState,
        target_id: str | None,
    ) -> tuple[str, _BrowserTabState]:
        resolved_target_id = target_id or state.active_target_id
        if resolved_target_id is not None:
            tab = state.tabs.get(resolved_target_id)
            if tab is not None and not tab.page.is_closed():
                state.active_target_id = resolved_target_id
                return resolved_target_id, tab
        for candidate_id, tab in state.tabs.items():
            if not tab.page.is_closed():
                state.active_target_id = candidate_id
                return candidate_id, tab
        page = await state.context.new_page()
        candidate_id = self._register_tab(state, page)
        state.active_target_id = candidate_id
        return candidate_id, state.tabs[candidate_id]

    def _register_tab(self, state: _BrowserProfileState, page: Page) -> str:
        target_id = uuid4().hex[:8]
        tab = _BrowserTabState(target_id=target_id, page=page)
        state.tabs[target_id] = tab
        page.on("console", lambda message: self._record_console(tab, message))
        page.on("dialog", lambda dialog: asyncio.create_task(self._handle_dialog(tab, dialog)))
        page.on("close", lambda _: self._remove_tab(state, target_id))
        return target_id

    @staticmethod
    def _record_console(tab: _BrowserTabState, message: ConsoleMessage) -> None:
        entry = {
            "level": message.type,
            "location": message.location,
            "text": message.text,
        }
        tab.console.append(entry)
        if len(tab.console) > _MAX_CONSOLE_ENTRIES:
            del tab.console[:-_MAX_CONSOLE_ENTRIES]

    async def _handle_dialog(self, tab: _BrowserTabState, dialog: Dialog) -> None:
        behavior = tab.pending_dialog
        if behavior is None:
            await dialog.dismiss()
            return
        tab.pending_dialog = None
        if behavior.get("accept"):
            await dialog.accept(str(behavior.get("promptText") or ""))
            return
        await dialog.dismiss()

    @staticmethod
    def _resolve_selector(tab: _BrowserTabState, ref_or_selector: str | None) -> str | None:
        if ref_or_selector is None:
            return None
        return tab.refs.get(ref_or_selector, ref_or_selector)

    def _next_output_path(self, extension: str) -> Path:
        return self._resolve_output_dir() / f"{uuid4().hex}.{extension}"

    def _resolve_output_dir(self) -> Path:
        """Return the directory used for browser artifacts."""
        if self._output_dir is not None:
            return self._output_dir

        context = get_tool_runtime_context()
        if context is None or context.storage_path is None:
            msg = "BrowserTools requires explicit output_dir or tool runtime context with storage_path"
            raise RuntimeError(msg)

        self._output_dir = (context.storage_path / "browser").resolve()
        self._output_dir.mkdir(parents=True, exist_ok=True)
        return self._output_dir

    @staticmethod
    def _remove_tab(state: _BrowserProfileState, target_id: str) -> None:
        state.tabs.pop(target_id, None)
        if state.active_target_id == target_id:
            state.active_target_id = next(iter(state.tabs.keys()), None)



#!/usr/bin/env python
"""Benchmark script to test Matrix server message throughput.

This script tests the raw performance of the Matrix protocol/server
by sending messages at various rates and measuring throughput.
"""

import asyncio
import logging
import os
import time
from dataclasses import dataclass
from typing import TYPE_CHECKING

import yaml

# Suppress all debug logs for cleaner benchmark output
os.environ["MINDROOM_LOG_LEVEL"] = "WARNING"  # Set before importing mindroom
logging.getLogger("mindroom").setLevel(logging.WARNING)
logging.getLogger("mindroom.matrix").setLevel(logging.WARNING)
logging.getLogger("mindroom.matrix.client").setLevel(logging.WARNING)

from mindroom.constants import matrix_state_file, resolve_primary_runtime_paths, runtime_matrix_homeserver  # noqa: E402
from mindroom.matrix.client import send_message  # noqa: E402
from mindroom.matrix.users import AgentMatrixUser, login_agent_user  # noqa: E402

# Suppress structlog output after import (mindroom uses structlog)
try:
    import structlog

    structlog.configure(
        wrapper_class=structlog.make_filtering_bound_logger(logging.WARNING),
    )
except ImportError:
    pass

if TYPE_CHECKING:
    import nio


@dataclass
class BenchmarkMetrics:
    """Metrics collected during benchmark."""

    messages_sent: int = 0
    messages_failed: int = 0
    start_time: float = 0
    end_time: float = 0

    @property
    def duration(self) -> float:
        """Calculate the duration of the benchmark."""
        return self.end_time - self.start_time

    @property
    def throughput(self) -> float:
        """Calculate messages per second throughput."""
        if self.duration == 0:
            return 0
        return self.messages_sent / self.duration

    @property
    def success_rate(self) -> float:
        """Calculate the percentage of successfully sent messages."""
        total = self.messages_sent + self.messages_failed
        if total == 0:
            return 0
        return self.messages_sent / total * 100


class MatrixBenchmark:
    """Benchmark Matrix server message throughput."""

    def __init__(self, homeserver: str | None = None) -> None:
        self.runtime_paths = resolve_primary_runtime_paths()
        self.homeserver = homeserver or runtime_matrix_homeserver(runtime_paths=self.runtime_paths)
        self.client: nio.AsyncClient | None = None
        self.room_id: str | None = None
        self.metrics = BenchmarkMetrics()
        self.agent_user: AgentMatrixUser | None = None

    async def setup(self) -> None:
        """Load credentials and setup client."""
        # Load credentials from matrix_state.yaml
        with matrix_state_file(self.runtime_paths).open() as f:
            data = yaml.safe_load(f)

        # Use the main user account
        user_data = data["accounts"]["agent_user"]
        username = user_data["username"]
        password = user_data["password"]

        # Use lobby room for testing
        self.room_id = data["rooms"]["lobby"]["room_id"]

        # Extract domain from homeserver
        domain = self.homeserver.split("://")[1] if "://" in self.homeserver else self.homeserver

        # Create AgentMatrixUser object
        self.agent_user = AgentMatrixUser(
            agent_name="benchmark",
            user_id=f"@{username}:{domain}",
            display_name="Benchmark Bot",
            password=password,
        )

    async def login(self) -> None:
        """Login to Matrix server."""
        assert self.agent_user is not None
        print(f"🔑 Logging in as {self.agent_user.user_id} to {self.homeserver}...")

        # Use the login utility from mindroom
        self.client = await login_agent_user(self.homeserver, self.agent_user, self.runtime_paths)
        print(f"✅ Logged in successfully as {self.agent_user.user_id}")

    async def send_message(self, content: str, batch_id: int, msg_id: int) -> bool:
        """Send a single message and return success status."""
        assert self.client is not None
        assert self.room_id is not None
        try:
            message_content = {"msgtype": "m.text", "body": f"[Benchmark B{batch_id:03d}M{msg_id:04d}] {content}"}

            # Use the send_message utility from mindroom
            event_id = await send_message(self.client, self.room_id, message_content)

            if event_id:
                return True
            print("❌ Failed to send message")
            return False  # noqa: TRY300

        except Exception as e:
            print(f"❌ Error sending message: {e}")
            return False

    async def run_burst_test(self, num_messages: int, concurrency: int = 1) -> BenchmarkMetrics:
        """Send messages in a burst with TRUE concurrency - all at once."""
        print(f"\n🚀 Burst Test: {num_messages} messages with TRUE concurrency={concurrency}")

        self.metrics = BenchmarkMetrics()

        # Create ALL tasks at once - no batching!
        tasks = []
        for i in range(num_messages):
            batch_id = i // max(1, concurrency)
            msg_id = i
            task = self.send_message(f"Burst test message {i + 1}/{num_messages}", batch_id, msg_id)
            tasks.append(task)

        # Execute ALL tasks concurrently
        print(f"    Sending {len(tasks)} messages simultaneously...")
        self.metrics.start_time = time.time()  # Start timer RIGHT before sending
        results = await asyncio.gather(*tasks, return_exceptions=True)

        # Count results
        for result in results:
            if isinstance(result, bool) and result:
                self.metrics.messages_sent += 1
            else:
                self.metrics.messages_failed += 1

        self.metrics.end_time = time.time()

        print(f"✅ Sent: {self.metrics.messages_sent}/{num_messages}")
        print(f"❌ Failed: {self.metrics.messages_failed}")
        print(f"⏱️  Duration: {self.metrics.duration:.2f}s")
        print(f"📊 Throughput: {self.metrics.throughput:.2f} msg/s")
        print(f"📈 Success Rate: {self.metrics.success_rate:.1f}%")

        return self.metrics

    async def run_sustained_test(self, duration_seconds: int, messages_per_second: float) -> BenchmarkMetrics:
        """Send messages at a sustained rate for a specified duration."""
        print(f"\n⏳ Sustained Test: {messages_per_second} msg/s for {duration_seconds}s")

        self.metrics = BenchmarkMetrics()
        self.metrics.start_time = time.time()

        interval = 1.0 / messages_per_second
        end_time = self.metrics.start_time + duration_seconds
        msg_count = 0

        while time.time() < end_time:
            loop_start = time.time()

            # Send message
            success = await self.send_message(f"Sustained test at {messages_per_second} msg/s", 0, msg_count)

            if success:
                self.metrics.messages_sent += 1
            else:
                self.metrics.messages_failed += 1

            msg_count += 1

            # Sleep to maintain rate
            elapsed = time.time() - loop_start
            if elapsed < interval:
                await asyncio.sleep(interval - elapsed)

        self.metrics.end_time = time.time()

        print(f"✅ Sent: {self.metrics.messages_sent}")
        print(f"❌ Failed: {self.metrics.messages_failed}")
        print(f"⏱️  Duration: {self.metrics.duration:.2f}s")
        print(f"📊 Actual Throughput: {self.metrics.throughput:.2f} msg/s")
        print(f"📈 Success Rate: {self.metrics.success_rate:.1f}%")

        return self.metrics

    async def run_ramp_test(self, max_rate: float, step_duration: int = 5) -> list[tuple[int, BenchmarkMetrics]]:
        """Gradually increase message rate to find maximum throughput."""
        print(f"\n📈 Ramp Test: 0 to {max_rate} msg/s")

        rates = [1, 2, 5, 10, 20, 30, 40, 50, 75, 100, 200]
        rates = [r for r in rates if r <= max_rate]

        results: list[tuple[int, BenchmarkMetrics]] = []

        for rate in rates:
            print(f"\n🔄 Testing {rate} msg/s...")
            metrics = await self.run_sustained_test(step_duration, rate)
            results.append((rate, metrics))

            # Stop if success rate drops below 90%
            if metrics.success_rate < 90:
                print(f"⚠️  Success rate dropped below 90% at {rate} msg/s")
                break

            # Brief pause between tests
            await asyncio.sleep(1)

        # Print summary
        print("\n📊 Ramp Test Summary:")
        print("Rate (msg/s) | Actual (msg/s) | Success Rate")
        print("-" * 45)
        for target_rate, metrics in results:
            print(f"{target_rate:11.1f} | {metrics.throughput:14.2f} | {metrics.success_rate:11.1f}%")

        return results

    async def cleanup(self) -> None:
        """Close client connection."""
        if self.client:
            await self.client.close()


async def main() -> None:
    """Run the benchmark suite."""
    benchmark = MatrixBenchmark()

    try:
        # Setup and login
        await benchmark.setup()
        await benchmark.login()

        print("\n" + "=" * 60)
        print("🎯 Matrix Server Throughput Benchmark")
        print("=" * 60)

        # Test 1: Burst test with different concurrency levels
        print("\n### BURST TESTS ###")
        for concurrency in [1, 5, 10, 20]:
            await benchmark.run_burst_test(100, concurrency=concurrency)
            await asyncio.sleep(2)  # Brief pause between tests

        # Test 2: Sustained rate test
        print("\n### SUSTAINED RATE TESTS ###")
        for rate in [5, 10, 20, 50]:
            await benchmark.run_sustained_test(duration_seconds=10, messages_per_second=rate)
            await asyncio.sleep(2)

        # Test 3: Ramp test to find maximum sustainable rate
        print("\n### RAMP TEST ###")
        await benchmark.run_ramp_test(max_rate=500, step_duration=5)

        print("\n" + "=" * 60)
        print("✅ Benchmark Complete!")
        print("=" * 60)

    finally:
        await benchmark.cleanup()


if __name__ == "__main__":
    asyncio.run(main())



"""Tests for syncing shared provider/bootstrap credentials from runtime env."""

import os
from pathlib import Path

import pytest

from mindroom import constants as constants_mod
from mindroom.credentials import SHARED_CREDENTIALS_PATH_ENV, CredentialsManager
from mindroom.credentials_sync import (
    _ENV_TO_SERVICE_MAP,
    get_api_key_for_provider,
    get_ollama_host,
    get_secret_from_env,
    sync_env_to_credentials,
)


def _runtime_paths(
    storage_root: Path,
    *,
    shared_credentials_dir: Path | None = None,
) -> constants_mod.RuntimePaths:
    config_path = storage_root / "config.yaml"
    config_path.write_text("agents: {}\nmodels: {}\nrouter:\n  model: default\n", encoding="utf-8")
    process_env = dict(os.environ)
    if shared_credentials_dir is not None:
        process_env[SHARED_CREDENTIALS_PATH_ENV] = str(shared_credentials_dir)
    return constants_mod.resolve_runtime_paths(
        config_path=config_path,
        storage_path=storage_root,
        process_env=process_env,
    )


class TestCredentialsSync:
    """Test the shared provider/bootstrap credential sync behavior."""

    @pytest.fixture
    def temp_credentials_dir(self, tmp_path: Path) -> Path:
        """Create a temporary credentials directory."""
        creds_dir = tmp_path / "credentials"
        creds_dir.mkdir()
        return creds_dir

    @pytest.fixture
    def credentials_manager(self, temp_credentials_dir: Path) -> CredentialsManager:
        """Create a CredentialsManager with a temporary directory."""
        return CredentialsManager(base_path=temp_credentials_dir)

    def test_sync_env_to_credentials_new_keys(
        self,
        temp_credentials_dir: Path,
        monkeypatch: pytest.MonkeyPatch,
    ) -> None:
        """Supported shared provider/bootstrap env values should seed credentials."""
        # Set shared provider/bootstrap env values.
        monkeypatch.setenv("OPENAI_API_KEY", "sk-test-openai-key")
        monkeypatch.setenv("ANTHROPIC_API_KEY", "sk-test-anthropic-key")
        monkeypatch.setenv("GOOGLE_API_KEY", "test-google-key")
        monkeypatch.setenv("OLLAMA_HOST", "http://test:11434")

        runtime_paths = _runtime_paths(
            temp_credentials_dir.parent,
            shared_credentials_dir=temp_credentials_dir,
        )

        # Run sync
        sync_env_to_credentials(runtime_paths=runtime_paths)

        # Verify files were created
        openai_file = temp_credentials_dir / "openai_credentials.json"
        anthropic_file = temp_credentials_dir / "anthropic_credentials.json"
        google_file = temp_credentials_dir / "google_credentials.json"
        ollama_file = temp_credentials_dir / "ollama_credentials.json"

        assert openai_file.exists()
        assert anthropic_file.exists()
        assert google_file.exists()
        assert ollama_file.exists()

        # Verify content
        cm = CredentialsManager(base_path=temp_credentials_dir)
        assert cm.get_api_key("openai") == "sk-test-openai-key"
        assert cm.get_api_key("anthropic") == "sk-test-anthropic-key"
        assert cm.get_api_key("google") == "test-google-key"

        # Verify source metadata is tracked
        openai_creds = cm.load_credentials("openai")
        assert openai_creds["_source"] == "env"

        ollama_creds = cm.load_credentials("ollama")
        assert ollama_creds["host"] == "http://test:11434"
        assert ollama_creds["_source"] == "env"

    def test_sync_env_does_not_overwrite_ui_credentials(
        self,
        temp_credentials_dir: Path,
        monkeypatch: pytest.MonkeyPatch,
    ) -> None:
        """Test that env sync does NOT overwrite UI-set credentials."""
        cm = CredentialsManager(base_path=temp_credentials_dir)
        cm.save_credentials("openai", {"api_key": "ui-set-key", "_source": "ui"})

        monkeypatch.setenv("OPENAI_API_KEY", "env-key")

        sync_env_to_credentials(
            runtime_paths=_runtime_paths(
                temp_credentials_dir.parent,
                shared_credentials_dir=temp_credentials_dir,
            ),
        )

        assert cm.get_api_key("openai") == "ui-set-key"

    def test_get_secret_from_env_resolves_relative_file_paths_from_config_dir(self, tmp_path: Path) -> None:
        """Relative *_FILE secret paths in the runtime `.env` should anchor to the config directory."""
        config_dir = tmp_path / "cfg"
        config_dir.mkdir(parents=True, exist_ok=True)
        config_path = config_dir / "config.yaml"
        config_path.write_text("agents: {}\nmodels: {}\nrouter:\n  model: default\n", encoding="utf-8")
        secret_file = config_dir / "secrets" / "openai.key"
        secret_file.parent.mkdir(parents=True, exist_ok=True)
        secret_file.write_text("sk-relative", encoding="utf-8")
        (config_dir / ".env").write_text("OPENAI_API_KEY_FILE=secrets/openai.key\n", encoding="utf-8")

        runtime_paths = constants_mod.resolve_runtime_paths(config_path=config_path, process_env={})

        assert get_secret_from_env("OPENAI_API_KEY", runtime_paths) == "sk-relative"

    def test_sync_env_does_not_overwrite_legacy_credentials(
        self,
        temp_credentials_dir: Path,
        monkeypatch: pytest.MonkeyPatch,
    ) -> None:
        """Test that env sync does NOT overwrite legacy credentials (no _source)."""
        cm = CredentialsManager(base_path=temp_credentials_dir)
        # Legacy credential without _source field
        cm.set_api_key("openai", "legacy-key")

        monkeypatch.setenv("OPENAI_API_KEY", "env-key")

        sync_env_to_credentials(
            runtime_paths=_runtime_paths(
                temp_credentials_dir.parent,
                shared_credentials_dir=temp_credentials_dir,
            ),
        )

        assert cm.get_api_key("openai") == "legacy-key"

    def test_sync_env_updates_env_sourced_credentials(
        self,
        temp_credentials_dir: Path,
        monkeypatch: pytest.MonkeyPatch,
    ) -> None:
        """Test that env sync DOES update env-sourced credentials."""
        cm = CredentialsManager(base_path=temp_credentials_dir)
        cm.save_credentials("openai", {"api_key": "old-env-key", "_source": "env"})

        monkeypatch.setenv("OPENAI_API_KEY", "new-env-key")

        sync_env_to_credentials(
            runtime_paths=_runtime_paths(
                temp_credentials_dir.parent,
                shared_credentials_dir=temp_credentials_dir,
            ),
        )

        assert cm.get_api_key("openai") == "new-env-key"

    def test_sync_env_to_credentials_skip_empty(
        self,
        temp_credentials_dir: Path,
        monkeypatch: pytest.MonkeyPatch,
    ) -> None:
        """Empty shared env values should be ignored."""
        # Set one valid and one empty shared env value.
        monkeypatch.setenv("OPENAI_API_KEY", "valid-key")
        monkeypatch.setenv("ANTHROPIC_API_KEY", "")

        cm = CredentialsManager(base_path=temp_credentials_dir)

        # Run sync
        sync_env_to_credentials(
            runtime_paths=_runtime_paths(
                temp_credentials_dir.parent,
                shared_credentials_dir=temp_credentials_dir,
            ),
        )

        # Verify only valid key was synced
        assert cm.get_api_key("openai") == "valid-key"
        assert cm.get_api_key("anthropic") is None

    def test_sync_env_seeds_github_private_from_github_token(
        self,
        temp_credentials_dir: Path,
        monkeypatch: pytest.MonkeyPatch,
    ) -> None:
        """GITHUB_TOKEN should seed github_private credentials for Git KB auth."""
        monkeypatch.setenv("GITHUB_TOKEN", "ghp-test-token")

        cm = CredentialsManager(base_path=temp_credentials_dir)
        sync_env_to_credentials(
            runtime_paths=_runtime_paths(
                temp_credentials_dir.parent,
                shared_credentials_dir=temp_credentials_dir,
            ),
        )

        github_private = cm.load_credentials("github_private")
        assert github_private == {
            "username": "x-access-token",
            "token": "ghp-test-token",
            "_source": "env",
        }

    def test_sync_env_does_not_overwrite_ui_github_private_credentials(
        self,
        temp_credentials_dir: Path,
        monkeypatch: pytest.MonkeyPatch,
    ) -> None:
        """UI-managed github_private credentials must not be overwritten by env sync."""
        cm = CredentialsManager(base_path=temp_credentials_dir)
        ui_value = "ui-value"
        cm.save_credentials(
            "github_private",
            {"username": "my-user", "token": ui_value, "_source": "ui"},
        )
        monkeypatch.setenv("GITHUB_TOKEN", "ghp-env-token")

        sync_env_to_credentials(
            runtime_paths=_runtime_paths(
                temp_credentials_dir.parent,
                shared_credentials_dir=temp_credentials_dir,
            ),
        )

        github_private = cm.load_credentials("github_private")
        assert github_private is not None
        assert github_private["token"] == ui_value
        assert github_private["_source"] == "ui"

    def test_get_api_key_for_provider(self, credentials_manager: CredentialsManager) -> None:
        """Test getting API key for different providers."""
        # Set up test data
        credentials_manager.set_api_key("openai", "test-openai-key")
        credentials_manager.set_api_key("google", "test-google-key")
        runtime_paths = _runtime_paths(
            credentials_manager.storage_root,
            shared_credentials_dir=credentials_manager.base_path,
        )

        # Test normal providers
        assert get_api_key_for_provider("openai", runtime_paths=runtime_paths) == "test-openai-key"
        assert get_api_key_for_provider("google", runtime_paths=runtime_paths) == "test-google-key"

        # Test gemini alias for google
        assert get_api_key_for_provider("gemini", runtime_paths=runtime_paths) == "test-google-key"

        # Test ollama returns None
        assert get_api_key_for_provider("ollama", runtime_paths=runtime_paths) is None

        # Test non-existent provider
        assert get_api_key_for_provider("anthropic", runtime_paths=runtime_paths) is None

    def test_get_ollama_host(self, credentials_manager: CredentialsManager) -> None:
        """Test getting Ollama host configuration."""
        # Test when no Ollama config exists
        runtime_paths = _runtime_paths(
            credentials_manager.storage_root,
            shared_credentials_dir=credentials_manager.base_path,
        )
        assert get_ollama_host(runtime_paths=runtime_paths) is None

        # Set Ollama host
        credentials_manager.save_credentials("ollama", {"host": "http://localhost:11434"})
        assert get_ollama_host(runtime_paths=runtime_paths) == "http://localhost:11434"

    def test_all_env_vars_mapped(self) -> None:
        """All supported shared provider/bootstrap env vars should be mapped."""
        expected_services = {
            "OPENAI_API_KEY": "openai",
            "ANTHROPIC_API_KEY": "anthropic",
            "GOOGLE_API_KEY": "google",
            "OPENROUTER_API_KEY": "openrouter",
            "DEEPSEEK_API_KEY": "deepseek",
            "CEREBRAS_API_KEY": "cerebras",
            "GROQ_API_KEY": "groq",
            "OLLAMA_HOST": "ollama",
        }

        assert expected_services == _ENV_TO_SERVICE_MAP

    def test_sync_idempotent(self, temp_credentials_dir: Path, monkeypatch: pytest.MonkeyPatch) -> None:
        """Test that running sync multiple times doesn't cause issues."""
        monkeypatch.setenv("OPENAI_API_KEY", "test-key")

        cm = CredentialsManager(base_path=temp_credentials_dir)

        # Run sync multiple times
        runtime_paths = _runtime_paths(
            temp_credentials_dir.parent,
            shared_credentials_dir=temp_credentials_dir,
        )
        sync_env_to_credentials(runtime_paths=runtime_paths)
        sync_env_to_credentials(runtime_paths=runtime_paths)
        sync_env_to_credentials(runtime_paths=runtime_paths)

        # Should still have the same value
        assert cm.get_api_key("openai") == "test-key"

        # Should only have one file
        openai_files = list(temp_credentials_dir.glob("openai_*.json"))
        assert len(openai_files) == 1



"""Tests for config auto-reload and room membership updates."""

from __future__ import annotations

import asyncio
import tempfile
from pathlib import Path
from unittest.mock import AsyncMock

import pytest

from mindroom.bot import AgentBot
from mindroom.config.agent import AgentConfig, CultureConfig, TeamConfig
from mindroom.config.main import Config
from mindroom.config.models import ModelConfig, RouterConfig
from mindroom.constants import ROUTER_AGENT_NAME
from mindroom.matrix.users import AgentMatrixUser
from mindroom.orchestration.config_updates import _get_changed_agents
from mindroom.orchestrator import MultiAgentOrchestrator
from tests.conftest import (
    TEST_PASSWORD,
    bind_runtime_paths,
    orchestrator_runtime_paths,
    runtime_paths_for,
    test_runtime_paths,
)


def _runtime_bound_config(config: Config, runtime_root: Path | None = None) -> Config:
    """Return a runtime-bound config for reload tests."""
    runtime_paths = test_runtime_paths(runtime_root or Path(tempfile.mkdtemp()))
    return bind_runtime_paths(config, runtime_paths)


def setup_test_bot(bot: AgentBot, mock_client: AsyncMock) -> None:
    """Helper to setup a test bot with required attributes."""
    bot.client = mock_client


def test_get_changed_agents_detects_culture_config_updates() -> None:
    """Agent restarts should trigger when their culture mode/assignment changes."""
    old_config = _runtime_bound_config(
        Config(
            agents={
                "agent1": AgentConfig(display_name="Agent 1"),
            },
            cultures={
                "engineering": CultureConfig(
                    description="Engineering standards",
                    agents=["agent1"],
                    mode="automatic",
                ),
            },
        ),
    )
    new_config = _runtime_bound_config(
        Config(
            agents={
                "agent1": AgentConfig(display_name="Agent 1"),
            },
            cultures={
                "engineering": CultureConfig(
                    description="Engineering standards",
                    agents=["agent1"],
                    mode="agentic",
                ),
            },
        ),
    )

    changed = _get_changed_agents(old_config, new_config, agent_bots={"agent1": AsyncMock()})
    assert changed == {"agent1"}


@pytest.fixture
def initial_config() -> Config:
    """Initial configuration with some agents and rooms."""
    return _runtime_bound_config(
        Config(
            agents={
                "agent1": AgentConfig(
                    display_name="Agent 1",
                    role="Test agent",
                    rooms=["room1", "room2"],
                ),
                "agent2": AgentConfig(
                    display_name="Agent 2",
                    role="Another test agent",
                    rooms=["room1"],
                ),
            },
            teams={
                "team1": TeamConfig(
                    display_name="Team 1",
                    role="Test team",
                    agents=["agent1", "agent2"],
                    rooms=["room3"],
                ),
            },
            models={
                "default": ModelConfig(
                    provider="ollama",
                    id="llama3.2",
                    host="http://localhost:11434",
                ),
            },
        ),
    )


@pytest.fixture
def updated_config() -> Config:
    """Updated configuration with changed room assignments."""
    return _runtime_bound_config(
        Config(
            agents={
                "agent1": AgentConfig(
                    display_name="Agent 1",
                    role="Test agent",
                    rooms=["room1", "room4"],  # Changed: removed room2, added room4
                ),
                "agent2": AgentConfig(
                    display_name="Agent 2",
                    role="Another test agent",
                    rooms=["room2", "room3"],  # Changed: removed room1, added room2 and room3
                ),
                "agent3": AgentConfig(  # New agent
                    display_name="Agent 3",
                    role="New agent",
                    rooms=["room5"],
                ),
            },
            teams={
                "team1": TeamConfig(
                    display_name="Team 1",
                    role="Test team",
                    agents=["agent1", "agent2", "agent3"],  # Added agent3
                    rooms=["room3", "room6"],  # Added room6
                ),
            },
            models={
                "default": ModelConfig(
                    provider="ollama",
                    id="llama3.2",
                    host="http://localhost:11434",
                ),
            },
        ),
    )


@pytest.fixture
def mock_agent_users() -> dict[str, AgentMatrixUser]:
    """Create mock agent users."""
    return {
        ROUTER_AGENT_NAME: AgentMatrixUser(
            agent_name=ROUTER_AGENT_NAME,
            user_id=f"@mindroom_{ROUTER_AGENT_NAME}:localhost",
            display_name="RouterAgent",
            password=TEST_PASSWORD,
        ),
        "agent1": AgentMatrixUser(
            agent_name="agent1",
            user_id="@mindroom_agent1:localhost",
            display_name="Agent 1",
            password=TEST_PASSWORD,
        ),
        "agent2": AgentMatrixUser(
            agent_name="agent2",
            user_id="@mindroom_agent2:localhost",
            display_name="Agent 2",
            password=TEST_PASSWORD,
        ),
        "agent3": AgentMatrixUser(
            agent_name="agent3",
            user_id="@mindroom_agent3:localhost",
            display_name="Agent 3",
            password=TEST_PASSWORD,
        ),
        "team1": AgentMatrixUser(
            agent_name="team1",
            user_id="@mindroom_team1:localhost",
            display_name="Team 1",
            password=TEST_PASSWORD,
        ),
    }


@pytest.mark.asyncio
async def test_agent_joins_new_rooms_on_config_reload(  # noqa: C901
    initial_config: Config,  # noqa: ARG001
    updated_config: Config,  # noqa: ARG001
    mock_agent_users: dict[str, AgentMatrixUser],
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
) -> None:
    """Test that agents join new rooms when their configuration is updated."""
    # Track room operations
    joined_rooms: dict[str, list[str]] = {}
    left_rooms: dict[str, list[str]] = {}

    async def mock_join_room(client: AsyncMock, room_id: str) -> bool:
        user_id = client.user_id
        if user_id not in joined_rooms:
            joined_rooms[user_id] = []
        joined_rooms[user_id].append(room_id)
        return True

    async def mock_leave_room(client: AsyncMock, room_id: str) -> bool:
        user_id = client.user_id
        if user_id not in left_rooms:
            left_rooms[user_id] = []
        left_rooms[user_id].append(room_id)
        return True

    monkeypatch.setattr("mindroom.bot.join_room", mock_join_room)
    monkeypatch.setattr("mindroom.matrix.rooms.leave_room", mock_leave_room)

    # Mock restore_scheduled_tasks
    async def mock_restore_scheduled_tasks(
        _client: AsyncMock,
        _room_id: str,
        _config: Config,
        _runtime_paths: object,
    ) -> int:
        return 0

    monkeypatch.setattr("mindroom.bot.restore_scheduled_tasks", mock_restore_scheduled_tasks)

    # Mock resolve_room_aliases
    def mock_resolve_room_aliases(aliases: list[str]) -> list[str]:
        return list(aliases)

    monkeypatch.setattr("mindroom.bot.resolve_room_aliases", mock_resolve_room_aliases)

    # Mock get_joined_rooms to simulate current room membership
    async def mock_get_joined_rooms(client: AsyncMock) -> list[str]:
        user_id = client.user_id
        if "agent1" in user_id:
            return ["room1", "room2"]  # agent1 is currently in room1 and room2
        if "agent2" in user_id:
            return ["room1"]  # agent2 is currently in room1
        if "team1" in user_id:
            return ["room3"]  # team1 is currently in room3
        if ROUTER_AGENT_NAME in user_id:
            return ["room1", "room2", "room3"]  # router is in all initial rooms
        return []

    monkeypatch.setattr("mindroom.bot.get_joined_rooms", mock_get_joined_rooms)

    # Create agent1 bot with initial config
    config = _runtime_bound_config(Config(router=RouterConfig(model="default")), tmp_path)
    agent1_bot = AgentBot(
        agent_user=mock_agent_users["agent1"],
        storage_path=tmp_path,
        config=config,
        runtime_paths=runtime_paths_for(config),
        rooms=["room1", "room2"],  # Initial rooms
    )
    mock_client = AsyncMock()
    mock_client.user_id = "@mindroom_agent1:localhost"
    setup_test_bot(agent1_bot, mock_client)

    # Update to new config rooms
    agent1_bot.rooms = ["room1", "room4"]  # New rooms: removed room2, added room4

    # Apply room updates
    await agent1_bot.join_configured_rooms()
    await agent1_bot.leave_unconfigured_rooms()

    # Verify agent1 joined room4 (new room)
    assert "room4" in joined_rooms.get("@mindroom_agent1:localhost", [])
    # Verify agent1 left room2 (no longer configured)
    assert "room2" in left_rooms.get("@mindroom_agent1:localhost", [])


@pytest.mark.asyncio
async def test_router_updates_rooms_on_config_reload(
    initial_config: Config,
    updated_config: Config,
    mock_agent_users: dict[str, AgentMatrixUser],
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
) -> None:
    """Test that the router updates its room list when agents/teams change their rooms."""
    # Track room operations
    joined_rooms: list[str] = []
    left_rooms: list[str] = []

    async def mock_join_room(_client: AsyncMock, room_id: str) -> bool:
        joined_rooms.append(room_id)
        return True

    async def mock_leave_room(_client: AsyncMock, room_id: str) -> bool:
        left_rooms.append(room_id)
        return True

    monkeypatch.setattr("mindroom.bot.join_room", mock_join_room)
    monkeypatch.setattr("mindroom.matrix.rooms.leave_room", mock_leave_room)

    # Mock restore_scheduled_tasks
    async def mock_restore_scheduled_tasks(
        _client: AsyncMock,
        _room_id: str,
        _config: Config,
        _runtime_paths: object,
    ) -> int:
        return 0

    monkeypatch.setattr("mindroom.bot.restore_scheduled_tasks", mock_restore_scheduled_tasks)

    # Mock resolve_room_aliases
    def mock_resolve_room_aliases(aliases: list[str]) -> list[str]:
        return list(aliases)

    monkeypatch.setattr("mindroom.bot.resolve_room_aliases", mock_resolve_room_aliases)

    # Mock get_joined_rooms to simulate current room membership
    async def mock_get_joined_rooms(_client: AsyncMock) -> list[str]:
        # Router is currently in initial config rooms
        return ["room1", "room2", "room3"]

    monkeypatch.setattr("mindroom.bot.get_joined_rooms", mock_get_joined_rooms)

    # Get initial router rooms
    initial_router_rooms = initial_config.get_all_configured_rooms()
    assert initial_router_rooms == {"room1", "room2", "room3"}

    # Get updated router rooms
    updated_router_rooms = updated_config.get_all_configured_rooms()
    assert updated_router_rooms == {"room1", "room2", "room3", "room4", "room5", "room6"}

    # Create router bot with updated config
    config = _runtime_bound_config(Config(router=RouterConfig(model="default")), tmp_path)
    router_bot = AgentBot(
        agent_user=mock_agent_users[ROUTER_AGENT_NAME],
        storage_path=tmp_path,
        config=config,
        runtime_paths=runtime_paths_for(config),
        rooms=list(updated_router_rooms),
    )
    mock_client = AsyncMock()
    mock_client.user_id = f"@mindroom_{ROUTER_AGENT_NAME}:localhost"
    setup_test_bot(router_bot, mock_client)

    # Apply room updates
    await router_bot.join_configured_rooms()
    await router_bot.leave_unconfigured_rooms()

    # Verify router joined new rooms
    for new_room in ["room4", "room5", "room6"]:
        assert new_room in joined_rooms

    # Router should not leave any rooms (all initial rooms still have agents)
    assert len(left_rooms) == 0


@pytest.mark.asyncio
async def test_new_agent_joins_rooms_on_config_reload(
    initial_config: Config,  # noqa: ARG001
    updated_config: Config,  # noqa: ARG001
    mock_agent_users: dict[str, AgentMatrixUser],
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
) -> None:
    """Test that new agents are created and join their configured rooms."""
    # Track room operations
    joined_rooms: dict[str, list[str]] = {}

    async def mock_ensure_all_agent_users(_homeserver: str) -> dict[str, AgentMatrixUser]:
        # Return both existing and new agent users
        return mock_agent_users

    monkeypatch.setattr("mindroom.matrix.users._ensure_all_agent_users", mock_ensure_all_agent_users)

    async def mock_join_room(client: AsyncMock, room_id: str) -> bool:
        user_id = client.user_id
        if user_id not in joined_rooms:
            joined_rooms[user_id] = []
        joined_rooms[user_id].append(room_id)
        return True

    monkeypatch.setattr("mindroom.bot.join_room", mock_join_room)

    # Mock restore_scheduled_tasks
    async def mock_restore_scheduled_tasks(
        _client: AsyncMock,
        _room_id: str,
        _config: Config,
        _runtime_paths: object,
    ) -> int:
        return 0

    monkeypatch.setattr("mindroom.bot.restore_scheduled_tasks", mock_restore_scheduled_tasks)

    # Mock resolve_room_aliases
    def mock_resolve_room_aliases(aliases: list[str]) -> list[str]:
        return list(aliases)

    monkeypatch.setattr("mindroom.bot.resolve_room_aliases", mock_resolve_room_aliases)

    # Mock get_joined_rooms
    async def mock_get_joined_rooms(_client: AsyncMock) -> list[str]:
        return []  # New agent has no rooms initially

    monkeypatch.setattr("mindroom.bot.get_joined_rooms", mock_get_joined_rooms)

    # Create agent3 bot (new agent in updated config)
    config = _runtime_bound_config(Config(router=RouterConfig(model="default")), tmp_path)
    agent3_bot = AgentBot(
        agent_user=mock_agent_users["agent3"],
        storage_path=tmp_path,
        config=config,
        runtime_paths=runtime_paths_for(config),
        rooms=["room5"],
    )
    mock_client = AsyncMock()
    mock_client.user_id = "@mindroom_agent3:localhost"
    setup_test_bot(agent3_bot, mock_client)

    # Apply room updates for new agent
    await agent3_bot.join_configured_rooms()

    # Verify agent3 joined its configured room
    assert "room5" in joined_rooms.get("@mindroom_agent3:localhost", [])


@pytest.mark.asyncio
async def test_team_room_changes_on_config_reload(
    initial_config: Config,  # noqa: ARG001
    updated_config: Config,  # noqa: ARG001
    mock_agent_users: dict[str, AgentMatrixUser],
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
) -> None:
    """Test that teams update their room memberships when configuration changes."""
    # Track room operations
    joined_rooms: dict[str, list[str]] = {}
    left_rooms: dict[str, list[str]] = {}

    async def mock_join_room(client: AsyncMock, room_id: str) -> bool:
        user_id = client.user_id
        if user_id not in joined_rooms:
            joined_rooms[user_id] = []
        joined_rooms[user_id].append(room_id)
        return True

    async def mock_leave_room(client: AsyncMock, room_id: str) -> bool:
        user_id = client.user_id
        if user_id not in left_rooms:
            left_rooms[user_id] = []
        left_rooms[user_id].append(room_id)
        return True

    monkeypatch.setattr("mindroom.bot.join_room", mock_join_room)
    monkeypatch.setattr("mindroom.matrix.rooms.leave_room", mock_leave_room)

    # Mock restore_scheduled_tasks
    async def mock_restore_scheduled_tasks(
        _client: AsyncMock,
        _room_id: str,
        _config: Config,
        _runtime_paths: object,
    ) -> int:
        return 0

    monkeypatch.setattr("mindroom.bot.restore_scheduled_tasks", mock_restore_scheduled_tasks)

    # Mock resolve_room_aliases
    def mock_resolve_room_aliases(aliases: list[str]) -> list[str]:
        return list(aliases)

    monkeypatch.setattr("mindroom.bot.resolve_room_aliases", mock_resolve_room_aliases)

    # Mock get_joined_rooms to simulate current room membership
    async def mock_get_joined_rooms(client: AsyncMock) -> list[str]:
        user_id = client.user_id
        if "team1" in user_id:
            return ["room3"]  # team1 is currently only in room3
        return []

    monkeypatch.setattr("mindroom.bot.get_joined_rooms", mock_get_joined_rooms)

    # Create team1 bot with updated config
    config = _runtime_bound_config(Config(router=RouterConfig(model="default")), tmp_path)
    team1_bot = AgentBot(
        agent_user=mock_agent_users["team1"],
        storage_path=tmp_path,
        config=config,
        runtime_paths=runtime_paths_for(config),
        rooms=["room3", "room6"],
    )
    mock_client = AsyncMock()
    mock_client.user_id = "@mindroom_team1:localhost"
    setup_test_bot(team1_bot, mock_client)

    # Apply room updates
    await team1_bot.join_configured_rooms()
    await team1_bot.leave_unconfigured_rooms()

    # Verify team1 joined room6 (new room)
    assert "room6" in joined_rooms.get("@mindroom_team1:localhost", [])
    # Team1 should not leave room3 (still configured)
    assert "room3" not in left_rooms.get("@mindroom_team1:localhost", [])


@pytest.mark.asyncio
@pytest.mark.requires_matrix  # This test requires a real Matrix server or extensive mocking
@pytest.mark.timeout(10)  # Add timeout to prevent hanging on real server connection
async def test_orchestrator_handles_config_reload(  # noqa: PLR0915
    initial_config: Config,
    updated_config: Config,
    mock_agent_users: dict[str, AgentMatrixUser],
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
) -> None:
    """Test that the orchestrator properly handles config reloads and updates all bots."""
    # Track config loads
    config_loads = [initial_config, updated_config]
    load_count = [0]

    def mock_load_config(_config_path: Path | None = None) -> Config:
        result = config_loads[min(load_count[0], len(config_loads) - 1)]
        load_count[0] += 1
        return result

    monkeypatch.setattr("mindroom.config.main.Config.from_yaml", mock_load_config)

    async def mock_ensure_all_agent_users(_homeserver: str) -> dict[str, AgentMatrixUser]:
        return mock_agent_users

    monkeypatch.setattr("mindroom.matrix.users._ensure_all_agent_users", mock_ensure_all_agent_users)

    def mock_resolve_room_aliases(aliases: list[str]) -> list[str]:
        return list(aliases)

    monkeypatch.setattr("mindroom.bot.resolve_room_aliases", mock_resolve_room_aliases)

    # Mock topic generation to avoid calling AI
    async def mock_generate_room_topic_ai(room_key: str, room_name: str, config: Config) -> str:  # noqa: ARG001
        return f"Test topic for {room_name}"

    monkeypatch.setattr("mindroom.topic_generator.generate_room_topic_ai", mock_generate_room_topic_ai)
    monkeypatch.setattr("mindroom.matrix.rooms.generate_room_topic_ai", mock_generate_room_topic_ai)

    # Create orchestrator
    # Mock start/sync at class level so newly created bots during update_config don't perform real login/sync
    # But we need to ensure client gets set when start() is called
    async def mock_start(self: AgentBot) -> None:
        """Mock start that sets a mock client."""
        self.client = AsyncMock()
        self.client.user_id = self.agent_user.user_id
        self.running = True

    monkeypatch.setattr("mindroom.bot.AgentBot.start", mock_start)
    monkeypatch.setattr("mindroom.bot.AgentBot.sync_forever", AsyncMock())
    monkeypatch.setattr("mindroom.bot.TeamBot.start", mock_start)
    monkeypatch.setattr("mindroom.bot.TeamBot.sync_forever", AsyncMock())
    monkeypatch.setattr("mindroom.orchestrator.MultiAgentOrchestrator._ensure_user_account", AsyncMock())
    monkeypatch.setattr("mindroom.orchestrator.MultiAgentOrchestrator._setup_rooms_and_memberships", AsyncMock())

    orchestrator = MultiAgentOrchestrator(runtime_paths=orchestrator_runtime_paths(tmp_path))

    # Initialize with initial config
    await orchestrator.initialize()

    # Verify initial state
    assert "agent1" in orchestrator.agent_bots
    assert "agent2" in orchestrator.agent_bots
    assert "agent3" not in orchestrator.agent_bots  # Not in initial config
    assert "team1" in orchestrator.agent_bots
    assert ROUTER_AGENT_NAME in orchestrator.agent_bots

    # Check initial room assignments
    assert set(orchestrator.agent_bots["agent1"].rooms) == {"room1", "room2"}
    assert set(orchestrator.agent_bots["agent2"].rooms) == {"room1"}
    assert set(orchestrator.agent_bots["team1"].rooms) == {"room3"}
    assert set(orchestrator.agent_bots[ROUTER_AGENT_NAME].rooms) == {"room1", "room2", "room3"}

    # Create a mock start method that initializes client
    async def mock_start_with_thread_manager(self: AgentBot) -> None:
        """Mock start that initializes client."""
        if not hasattr(self, "client") or self.client is None:
            self.client = AsyncMock()
            self.client.user_id = self.agent_user.user_id

    # Patch AgentBot.start and TeamBot.start to use our mock
    monkeypatch.setattr("mindroom.bot.AgentBot.start", mock_start_with_thread_manager)
    monkeypatch.setattr("mindroom.bot.TeamBot.start", mock_start_with_thread_manager)

    # Mock bot operations for update
    for bot in orchestrator.agent_bots.values():
        monkeypatch.setattr(bot, "stop", AsyncMock())
        monkeypatch.setattr(bot, "start", mock_start_with_thread_manager)
        monkeypatch.setattr(bot, "ensure_user_account", AsyncMock())
        monkeypatch.setattr(bot, "sync_forever", AsyncMock(side_effect=asyncio.CancelledError()))

    # Update config
    updated = await orchestrator.update_config()
    assert updated  # Should return True since config changed

    # Verify updated state
    assert "agent1" in orchestrator.agent_bots
    assert "agent2" in orchestrator.agent_bots
    assert "agent3" in orchestrator.agent_bots  # New agent added
    assert "team1" in orchestrator.agent_bots
    assert ROUTER_AGENT_NAME in orchestrator.agent_bots

    # Check updated room assignments
    assert set(orchestrator.agent_bots["agent1"].rooms) == {"room1", "room4"}
    assert set(orchestrator.agent_bots["agent2"].rooms) == {"room2", "room3"}
    assert set(orchestrator.agent_bots["agent3"].rooms) == {"room5"}
    assert set(orchestrator.agent_bots["team1"].rooms) == {"room3", "room6"}
    assert set(orchestrator.agent_bots[ROUTER_AGENT_NAME].rooms) == {
        "room1",
        "room2",
        "room3",
        "room4",
        "room5",
        "room6",
    }


@pytest.mark.asyncio
async def test_room_membership_state_after_config_update(  # noqa: C901, PLR0915
    initial_config: Config,  # noqa: ARG001
    updated_config: Config,  # noqa: ARG001
    mock_agent_users: dict[str, AgentMatrixUser],
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
) -> None:
    """Test that room membership state is correct after config updates."""
    # Simulate room membership state
    room_memberships = {
        "room1": [
            "@mindroom_agent1:localhost",
            "@mindroom_agent2:localhost",
            f"@mindroom_{ROUTER_AGENT_NAME}:localhost",
        ],
        "room2": ["@mindroom_agent1:localhost", f"@mindroom_{ROUTER_AGENT_NAME}:localhost"],
        "room3": ["@mindroom_team1:localhost", f"@mindroom_{ROUTER_AGENT_NAME}:localhost"],
    }

    def update_room_membership(user_id: str, room_id: str, action: str) -> None:
        """Update simulated room membership."""
        if action == "join":
            if room_id not in room_memberships:
                room_memberships[room_id] = []
            if user_id not in room_memberships[room_id]:
                room_memberships[room_id].append(user_id)
        elif action == "leave":
            if room_id in room_memberships and user_id in room_memberships[room_id]:
                room_memberships[room_id].remove(user_id)

    async def mock_join_room(client: AsyncMock, room_id: str) -> bool:
        update_room_membership(client.user_id, room_id, "join")
        return True

    async def mock_leave_room(client: AsyncMock, room_id: str) -> bool:
        update_room_membership(client.user_id, room_id, "leave")
        return True

    monkeypatch.setattr("mindroom.bot.join_room", mock_join_room)
    monkeypatch.setattr("mindroom.matrix.rooms.leave_room", mock_leave_room)

    # Mock restore_scheduled_tasks
    async def mock_restore_scheduled_tasks(
        _client: AsyncMock,
        _room_id: str,
        _config: Config,
        _runtime_paths: object,
    ) -> int:
        return 0

    monkeypatch.setattr("mindroom.bot.restore_scheduled_tasks", mock_restore_scheduled_tasks)

    # Mock resolve_room_aliases
    def mock_resolve_room_aliases(aliases: list[str]) -> list[str]:
        return list(aliases)

    monkeypatch.setattr("mindroom.bot.resolve_room_aliases", mock_resolve_room_aliases)

    # Mock get_joined_rooms based on room_memberships
    async def mock_get_joined_rooms(client: AsyncMock) -> list[str]:
        user_id = client.user_id
        rooms = []
        for room_id, members in room_memberships.items():
            if user_id in members:
                rooms.append(room_id)
        return rooms

    monkeypatch.setattr("mindroom.bot.get_joined_rooms", mock_get_joined_rooms)

    # Apply config updates for each bot
    bots_config = {
        "@mindroom_agent1:localhost": {"old": ["room1", "room2"], "new": ["room1", "room4"]},
        "@mindroom_agent2:localhost": {"old": ["room1"], "new": ["room2", "room3"]},
        "@mindroom_agent3:localhost": {"old": [], "new": ["room5"]},
        "@mindroom_team1:localhost": {"old": ["room3"], "new": ["room3", "room6"]},
        f"@mindroom_{ROUTER_AGENT_NAME}:localhost": {
            "old": ["room1", "room2", "room3"],
            "new": ["room1", "room2", "room3", "room4", "room5", "room6"],
        },
    }

    # Simulate config update for each bot
    for user_id, bot_config in bots_config.items():
        mock_client = AsyncMock()
        mock_client.user_id = user_id

        # Determine which agent this is
        if "agent1" in user_id:
            agent_user = mock_agent_users["agent1"]
        elif "agent2" in user_id:
            agent_user = mock_agent_users["agent2"]
        elif "agent3" in user_id:
            agent_user = mock_agent_users["agent3"]
        elif "team1" in user_id:
            agent_user = mock_agent_users["team1"]
        else:
            agent_user = mock_agent_users[ROUTER_AGENT_NAME]

        config = _runtime_bound_config(Config(router=RouterConfig(model="default")), tmp_path)

        bot = AgentBot(
            agent_user=agent_user,
            storage_path=tmp_path,
            config=config,
            runtime_paths=runtime_paths_for(config),
            rooms=bot_config["new"],
        )
        setup_test_bot(bot, mock_client)

        await bot.join_configured_rooms()
        await bot.leave_unconfigured_rooms()

    # Verify final room membership state
    assert set(room_memberships.get("room1", [])) == {
        "@mindroom_agent1:localhost",
        f"@mindroom_{ROUTER_AGENT_NAME}:localhost",
    }
    assert set(room_memberships.get("room2", [])) == {
        "@mindroom_agent2:localhost",
        f"@mindroom_{ROUTER_AGENT_NAME}:localhost",
    }
    assert set(room_memberships.get("room3", [])) == {
        "@mindroom_agent2:localhost",
        "@mindroom_team1:localhost",
        f"@mindroom_{ROUTER_AGENT_NAME}:localhost",
    }
    assert set(room_memberships.get("room4", [])) == {
        "@mindroom_agent1:localhost",
        f"@mindroom_{ROUTER_AGENT_NAME}:localhost",
    }
    assert set(room_memberships.get("room5", [])) == {
        "@mindroom_agent3:localhost",
        f"@mindroom_{ROUTER_AGENT_NAME}:localhost",
    }
    assert set(room_memberships.get("room6", [])) == {
        "@mindroom_team1:localhost",
        f"@mindroom_{ROUTER_AGENT_NAME}:localhost",
    }
